var latestKnownSettings = {};
var loadingNewData = false;
var indexFromWebSocket = false;
var acceptingRemoteSettings = true;
var nbsp = " ";
var incrementValue = document.getElementById("incrementValue");
var preview_frame = document.getElementById("preview_frame");
var preview_url =
  window.location.origin +
  window.location.pathname.replace(/control\/?$/, "display/?test=1") +
  window.location.hash;
preview_frame?.setAttribute("src", preview_url);

setTimeout(
  () =>
    sendThroughWebSocket(
      JSON.stringify({
        type: "SettingsRequest",
      })
    ),
  50
);

function handleMessage(message) {
  if (message.type == "MessageBroadcast")
    try {
      handleBroadcastedMessage(JSON.parse(message.message));
    } catch (e) {
      console.error(message);
      console.error(e);
    }
}

function handleBroadcastedMessage(message) {
  if (message.type == "SettingsRequest") {
    sendThroughWebSocket(
      JSON.stringify({
        type: "Settings",
        settings: getSettings(),
      })
    );
  } else if (
    message.type == "NewSettings" ||
    (message.type == "Settings" && acceptingRemoteSettings)
  ) {
    acceptingRemoteSettings = true;
    applyLoadedSettings(message.settings);
    acceptingRemoteSettings = false;
  }
  console.log(message);
}

function downloadTextFile(filename, text) {
  var element = document.createElement("a");
  element.setAttribute(
    "href",
    "data:text/plain;charset=utf-8," + encodeURIComponent(text)
  );
  element.setAttribute("download", filename);
  element.style.display = "none";
  document.body.appendChild(element);
  element.click();
  document.body.removeChild(element);
}

function getSettings() {
  var settings = JSON.parse(JSON.stringify(latestKnownSettings));
  for (var form of Array.from(document.querySelectorAll("form.form")))
    for (var keyPair of Array.from(new FormData(form).entries()))
      settings[keyPair[0]] = keyPair[1];
  return settings;
}

function saveSettings() {
  var token = window.location.hash.substr(1);
  downloadTextFile(
    `coop-table-displayer_settings_${token}.json`,
    JSON.stringify(getSettings())
  );
}

function loadSettings(input) {
  var file = input.files[0];
  input.value = null;
  if (file) {
    var fr = new FileReader();
    fr.onload = () => {
      var text = fr.result;
      applyLoadedSettings(JSON.parse(text));
    };
    fr.readAsText(file);
  }
}

function applyLoadedSettings(settings) {
  loadingNewData = true;
  applyLoadedSettingsToTable(settings);
  applyOptionsToSection(settings);
  for (var setting of Object.entries(settings)) {
    var key = setting[0];
    var value = setting[1];
    latestKnownSettings[key] = value;
    var field = document.querySelector(`[name="${key}"]:not(.unsynced)`);
    if (!!field) {
      if (field.type == "radio" || field.type == "checkbox") {
        field = document.querySelector(`[name="${key}"][value="${value}"]`);
        field.checked = true;
      } else {
        field.value = value;
      }
    }
  }
  loadingNewData = false;
  if (!acceptingRemoteSettings) broadcastChanges(); // only once
}

function applyLoadedSettingsToTable(settings) {
  var managedtbl = document.querySelector("#managedtbl");
  if (managedtbl === null) {
    // silently ignore table being absent
  } else if (
    (+settings.tableRows || 0) <= 0 ||
    (+settings.tableCols || 0) <= 0
  ) {
    while (!!managedtbl.lastChild) managedtbl.removeChild(managedtbl.lastChild);
  } else if (
    (+settings.tableRows || 0) !== (+latestKnownSettings.tableRows || 0) ||
    (+settings.tableCols || 0) !== (+latestKnownSettings.tableCols || 0)
  ) {
    while (!!managedtbl.lastChild) managedtbl.removeChild(managedtbl.lastChild);
    var tbody = document.createElement("tbody");
    tbody.classList.add("w-100");
    for (var rowid = 0; rowid < +settings.tableRows; rowid++) {
      var tr = document.createElement("tr");
      tr.classList.add("flex");
      for (var colid = 0; colid < +settings.tableCols; colid++) {
        var td = document.createElement("td");
        td.classList.add("flex");
        var inp = document.createElement("input");
        inp.setAttribute("name", `tablecell_${rowid}_${colid}`);
        inp.setAttribute("style", "min-width: 1rem; width: 100%;");
        inp.setAttribute("onchange", "broadcastChanges()");
        inp.setAttribute("onkeyup", "broadcastChanges()");
        inp.setAttribute("onkeydown", "broadcastChanges()");
        td.appendChild(inp);
        tr.appendChild(td);
      }
      tbody.appendChild(tr);
    }
    managedtbl.appendChild(tbody);
  } else {
    // let the updater check value field by field
  }
}

function applyOptionsToSection(settings) {
  var perColumnSettings = document.querySelector("#perColumnSettings");
  if (perColumnSettings === null) {
    // silently ignore table being absent
  } else if ((+settings.tableCols || 0) <= 0) {
    while (!!perColumnSettings.lastChild)
      perColumnSettings.removeChild(perColumnSettings.lastChild);
  } else if (
    (+settings.tableCols || 0) !== (+latestKnownSettings.tableCols || 0)
  ) {
    while (!!perColumnSettings.lastChild)
      perColumnSettings.removeChild(perColumnSettings.lastChild);
    for (var i = 0; i < +settings.tableCols; i++) {
      var h3 = document.createElement("h3");
      h3.appendChild(document.createTextNode(`Column #${i + 1}`));
      perColumnSettings.appendChild(h3);
      appendLabeledInputToElement(
        perColumnSettings,
        `Font Family Column #${i + 1}`,
        `fontFamilyCol${i + 1}`,
        "",
        { placeholder: `inherit (${settings.fontFamily})` }
      );
      appendLabeledInputToElement(
        perColumnSettings,
        `Font Weight Column #${i + 1}`,
        `fontWeightCol${i + 1}`,
        "",
        { placeholder: `inherit (${settings.baseFontWeight})` }
      );
      appendLabeledInputToElement(
        perColumnSettings,
        `Font Color Column #${i + 1}`,
        `colorCol${i + 1}`,
        "",
        { placeholder: `inherit (${settings.fontColor})` }
      );
      appendLabeledInputToElement(
        perColumnSettings,
        `Font Size Column #${i + 1}`,
        `sizeCol${i + 1}`,
        "1",
        { type: "number", step: "0.01" }
      );
      appendLabeledInputToElement(
        perColumnSettings,
        `Background Color Column #${i + 1}`,
        `backgroundColorCol${i + 1}`,
        "transparent"
      );
      appendLabeledInputToElement(
        perColumnSettings,
        `Text width weight Column #${i + 1}`,
        `flexWeightCol${i + 1}`,
        "0",
        { type: "number", step: "1" }
      );
      appendLabeledInputToElement(
        perColumnSettings,
        `Background Margin Column #${i + 1}`,
        `backgroundMarginCol${i + 1}`,
        "0",
        { type: "number", step: "0.01" }
      );
      appendLabeledInputToElement(
        perColumnSettings,
        `Background Padding Column #${i + 1}`,
        `backgroundPaddingCol${i + 1}`,
        "0",
        { type: "number", step: "0.01" }
      );
      appendLabeledInputToElement(
        perColumnSettings,
        `Background Border Radius Column #${i + 1}`,
        `backgroundBorderRadiusCol${i + 1}`,
        "0",
        { type: "number", step: "0.01" }
      );
    }
  }
}

function appendLabeledInputToElement(parent, dname, name, value, extras) {
  extras = extras ?? {};
  var label = document.createElement("label");
  label.setAttribute("for", name);
  label.setAttribute("class", "f6 b db mb1 mt3 sans-serif mid-gray");
  label.appendChild(document.createTextNode(dname));
  var inp = document.createElement("input");
  inp.setAttribute("id", name);
  inp.setAttribute("name", name);
  inp.setAttribute("type", extras.type ?? "text");
  inp.setAttribute(
    "class",
    extras.class ?? "w-100 f5 pv3 ph3 bg-light-gray bn"
  );
  inp.setAttribute("onkeyup", extras.onkeyup ?? "broadcastChanges()");
  inp.setAttribute("onkeydown", extras.onkeydown ?? "broadcastChanges()");
  inp.setAttribute("onchange", extras.onchange ?? "broadcastChanges()");
  inp.setAttribute("value", value);
  for (var pair of Object.entries(extras)) inp.setAttribute(pair[0], pair[1]);
  parent.appendChild(label);
  parent.appendChild(inp);
}

function broadcastChanges() {
  if (!loadingNewData)
    sendThroughWebSocket(
      JSON.stringify({
        type: "NewSettings",
        settings: getSettings(),
      })
    );
}

function updateCriticalPath() {
  applyLoadedSettings(getSettings());
}

function compareDisplayContent() {
  var nothingChanged = true;
  for (let lazyInput of Array.from(
    document.getElementsByClassName("lazyInput")
  )) {
    let dcStore = lazyInput;
    let dcInput = document.querySelector(`[name="${lazyInput.name}Visible"]`);
    if (dcStore.value != dcInput.value) nothingChanged = false;
  }
  var btn = document.getElementById("btn-update-internal");
  if (!nothingChanged) btn.style.display = "block";
  else btn.style.display = "none";
  broadcastChanges();
}

function updateInternalDisplayContent() {
  for (let lazyInput of Array.from(
    document.getElementsByClassName("lazyInput")
  )) {
    let dcStore = lazyInput;
    let dcInput = document.querySelector(`[name="${lazyInput.name}Visible"]`);
    dcStore.value = dcInput.value;
  }
  compareDisplayContent();
}

var escapeHtml = (unsafe) =>
  unsafe
    .replace(/&/g, "&amp;")
    .replace(/</g, "&lt;")
    .replace(/>/g, "&gt;")
    .replace(/"/g, "&quot;")
    .replace(/'/g, "&#039;");
