var knownSettings = {};

var main = document.createElement("main");
document.body.appendChild(main);

var unorderedlist = document.createElement("ul");
main.appendChild(unorderedlist);

var nextStyle = document.getElementById("nextStyle");

setTimeout(
  () =>
    sendThroughWebSocket(
      JSON.stringify({
        type: "SettingsRequest",
      })
    ),
  50
);

function handleMessage(message) {
  if (message.type == "MessageBroadcast")
    try {
      handleBroadcastedMessage(JSON.parse(message.message));
    } catch (e) {
      console.error(message);
      console.error(e);
    }
  if (message.type == "YouAre")
    setTimeout(
      () =>
        sendThroughWebSocket(
          JSON.stringify({
            type: "SettingsRequest",
          })
        ),
      10
    );
  if (message.type == "ClientConnected") {
    if (!knownSettings)
      setTimeout(
        () =>
          sendThroughWebSocket(
            JSON.stringify({
              type: "SettingsRequest",
            })
          ),
        10
      );
  }
}

function handleBroadcastedMessage(message) {
  if (message.type == "Settings" || message.type == "NewSettings") {
    Object.entries(message.settings).forEach(
      ([k, v]) => (knownSettings[k] = v)
    );
    configureForSettings();
  } else if (message.type == "SettingsRequest") {
    if (knownSettings)
      sendThroughWebSocket(
        JSON.stringify({
          type: "Settings",
          settings: knownSettings,
        })
      );
  }
  console.log(message);
}

var escapeHtml = (unsafe) =>
  unsafe
    .replace(/&/g, "&amp;")
    .replace(/</g, "&lt;")
    .replace(/>/g, "&gt;")
    .replace(/"/g, "&quot;")
    .replace(/'/g, "&#039;");

class AsyncLazyEval {
  /**
   * @param {Promise<Object>} promise
   */
  constructor(promise) {
    this._promise = promise;
    this._evaluating = false;
    this._evaluated = false;
    this._thrown = false;
    this._value = null;
  }

  get eval() {
    if (this._evaluated)
      return new Promise((resolve, reject) => {
        if (!this._thrown) resolve(this._value);
        else reject(this._value);
      });
    else if (!this._evaluating) {
      this._evaluating = true;
      return new Promise((resolve, reject) => {
        this._promise
          .then((r) => {
            this._evaluating = false;
            this._evaluated = true;
            this._thrown = false;
            this._value = r;
            resolve(r);
          })
          .catch((r) => {
            this._evaluating = false;
            this._evaluated = true;
            this._thrown = true;
            this._value = r;
            reject(r);
          });
      });
    } else {
      let p = new Promise(async (resolve, reject) => {
        while (!this._evaluated) await asyncSleep(100);
        if (!this._thrown) resolve(this._value);
        else reject(this._value);
      });
      return p;
    }
  }
}

var conventionJsonCache = {};

/**
 * @param {number} ms
 * @returns {Promise<void>}
 */
function asyncSleep(ms) {
  return new Promise((resolve) => setTimeout(resolve, ms));
}

setTimeout(async () => {
  while (1) {
    try {
      await configureForSettings();
    } catch (e) {}
    await asyncSleep(20 * 1000);
  }
}, 50);

/**
 * @param {string} url
 */
async function ensureCacheUpdated(url) {
  var cached = conventionJsonCache[knownSettings.scheduleUrl];
  var time = new Date().getTime();
  if (!cached || (!!cached && time > cached.taken + 5 * 60 * 1000))
    cached = conventionJsonCache[knownSettings.scheduleUrl] = {
      taken: new Date().getTime(),
      data: new AsyncLazyEval(
        (async () => {
          if (url === undefined) throw new Error("Url cannot be undefined");
          const response = await fetch(
            `${url}${url?.includes("?") ? "&" : "?"}nonce=${time}`
          );
          const jsonData = await response.json();
          return jsonData;
        })()
      ),
    };
  try {
    return await cached.data.eval;
  } catch (e) {
    console.error(e);
    return undefined;
  }
}

/**
 * @param {number} n
 * @returns {number[]}
 */
function countUntil(n) {
  return Array(n)
    .fill(0)
    .map((_, i) => i);
}

function lpad(text, size, fill) {
  var t = String(text);
  while (t.length < size) t = `${fill}${t}`;
  return t;
}

function rangeOverlap(this_start, this_finish, that_start, that_finish) {
  return (
    (this_start >= that_start && this_start < that_finish) ||
    (this_finish > that_start && this_finish < that_finish) ||
    (that_start >= this_start && that_start < this_finish) ||
    (that_finish > this_start && that_finish < this_finish)
  );
}

var latestKnownSettings = {};

function applyLoadedSettingsToTable(settings) {
  var managedtbl = document.querySelector("#managedtbl");
  if (managedtbl === null) {
    // silently ignore table being absent
  } else if (
    (+settings.tableRows || 0) <= 0 ||
    (+settings.tableCols || 0) <= 0
  ) {
    while (!!managedtbl.lastChild) managedtbl.removeChild(managedtbl.lastChild);
  } else if (
    (+settings.tableRows || 0) !== (+latestKnownSettings.tableRows || 0) ||
    (+settings.tableCols || 0) !== (+latestKnownSettings.tableCols || 0)
  ) {
    while (!!managedtbl.lastChild) managedtbl.removeChild(managedtbl.lastChild);
    var tbody = document.createElement("tbody");
    for (var rowid = 0; rowid < +settings.tableRows; rowid++) {
      var tr = document.createElement("tr");
      for (var colid = 0; colid < +settings.tableCols; colid++) {
        var td = document.createElement("td");
        td.setAttribute("id", `tablecell_${rowid}_${colid}`);
        td.setAttribute("class", `tablecol_${colid}`);
        tr.appendChild(td);
      }
      tbody.appendChild(tr);
    }
    managedtbl.appendChild(tbody);
  } else {
    // let the updater check value field by field
  }
}

async function configureForSettings() {
  if (!!knownSettings) {
    var managedtbl = document.querySelector("#managedtbl");
    if (!managedtbl) {
      managedtbl = document.createElement("table");
      managedtbl.setAttribute("id", "managedtbl");
      document.body.appendChild(managedtbl);
    }
    applyLoadedSettingsToTable(knownSettings);
    if (
      (+knownSettings.tableRows || 0) > 0 ||
      (+knownSettings.tableCols || 0) > 0
    ) {
      for (var rowid = 0; rowid < +knownSettings.tableRows; rowid++) {
        for (var colid = 0; colid < +knownSettings.tableCols; colid++) {
          var cell = document.querySelector(`#tablecell_${rowid}_${colid}`);
          if (cell)
            cell.innerText = knownSettings[`tablecell_${rowid}_${colid}`] ?? "";
        }
      }
    }
    for (var pair of Object.entries(knownSettings))
      latestKnownSettings[pair[0]] = pair[1];
    nextStyle.innerHTML =
      `
      html, body {
        margin: 0;
        padding: 0;
      }
      body > table {
        margin: 0;
        padding: 0;
        width: 100%;
        font-family: ${knownSettings.fontFamily ?? "sans-serif"};
        font-weight: ${knownSettings.baseFontWeight ?? 500};
        font-size: ${knownSettings.baseFontSize ?? 1.2}rem;
        color: ${knownSettings.fontColor ?? "#000000"};
      }
      body > table > tbody > tr {
        display: flex;
        margin: ${knownSettings.cardMargin ?? 1}em;
        padding: ${knownSettings.cardPadding ?? 0.5}em;
        border-radius: ${knownSettings.cardBorderRadius ?? 0.25}em;
        background: ${knownSettings.cardColor ?? "#00000080"};
        overflow: hidden;
        white-space: nowrap;
        overflow: hidden;
        text-overflow: ellipsis;
        justify-content: space-between;
        align-items: center;
      }
      body > table > tbody > tr > td {
        display: flex;
        justify-content: stretch;
        align-items: center;
      }
    `.trim() +
      countUntil(+knownSettings.tableCols)
        .map((i) =>
          `
          body > table > tbody > tr > td.tablecol_${i} {
            font-family: ${
              (knownSettings[`fontFamilyCol${i + 1}`] ||
                knownSettings.fontFamily) ??
              "sans-serif"
            };
            font-weight: ${
              (knownSettings[`fontWeightCol${i + 1}`] ||
                knownSettings.baseFontWeight) ??
              300
            };
            font-size: ${knownSettings[`sizeCol${i + 1}`] ?? 1}em;
            overflow: hidden;
            white-space: nowrap;
            overflow: hidden;
            text-overflow: ellipsis;
            flex-grow: ${knownSettings[`flexWeightCol${i + 1}`] ?? 0};
            margin: ${knownSettings[`backgroundMarginCol${i + 1}`] ?? 0}em;
            padding: ${knownSettings[`backgroundPaddingCol${i + 1}`] ?? 0}em;
            border-radius: ${
              knownSettings[`backgroundBorderRadiusCol${i + 1}`] ?? 0
            }em;
            background: ${
              knownSettings[`backgroundColorCol${i + 1}`] ?? "transparent"
            };
            color: ${
              (knownSettings[`colorCol${i + 1}`] || knownSettings.fontColor) ??
              "#000000"
            };
          }
          `.trim()
        )
        .join("\n");
  }
}

var incrementDisplayBuffer = [];
function displayIncrement(formatted) {
  if (!!formatted) {
    var emptyBuffer = incrementDisplayBuffer.length == 0;
    incrementDisplayBuffer.push(formatted);
    if (emptyBuffer) processIncrementDisplayBuffer();
  }
}

function processIncrementDisplayBuffer() {
  if (incrementDisplayBuffer.length > 0) {
    var toDisplay = incrementDisplayBuffer[0];
    increment.innerText = toDisplay;
    increment.classList.remove("transparent");
    setTimeout(() => {
      setTimeout(() => {
        increment.classList.add("transparent");
        setTimeout(() => {
          increment.innerText = "";
          incrementDisplayBuffer.shift();
          processIncrementDisplayBuffer();
        }, knownSettings.transitionsDuration);
      }, knownSettings.diffDuration * 1000);
    }, knownSettings.transitionsDuration);
  }
}
